#include <sys/types.h>
#include <sys/wait.h>
#include <time.h>
#include <malloc.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <stdlib.h>
#include <ctype.h>

#include "util.h"
#include "object.h"
#include "classmember.h"
#include "directory.h"
#include "archive.h"
#include "source.h"
#include "file.h"
#include "class.h"
#include "variable.h"
#include "function.h"
#include "edge.h"
#include "asxo/asxo.h"
#include "cx/cx.h"

// #define SANITY

static char	*cwdP = 0;
static char *cwd1P = 0;
static char *g_program_nameP = 0;
static int	g_ret            = 0;

static char	cwd1[2048];
#define No_variables		0
#define Global_variables	1
#define Static_variables	2
#define Static_function_variables	3

static int	show_variables  = Global_variables;
static int	no_function_calls= 0;

#define SUFFIX_a    0
#define SUFFIX_so   1
#define SUFFIX_o 	2
#define SUFFIX_S 	3
#define SUFFIX_c 	4
#define SUFFIX_C 	5
#define SUFFIX_cc 	6
#define SUFFIX_cxx	7
#define SUFFIX_cpp	8
#define SUFFIX_cpp1 9
#define SUFFIX_none 10

/* Must agree with SUFFIX_* */

static char *suffixes[] =
{
	"a",
	"so",
	"o",
	"s",
	"c",
	"C",
	"cc",
	"cxx",
	"cpp",
	"c++"
};

static int force[sizeof(suffixes)/sizeof(*suffixes)] = {0};

static int use_stdin = 0;

/* Convert the pathP to an absolute path.  If cdP != null then the path
 * is with respect to cwdP/cdP because we did a cd into cdP before
 * building the assember
 */

static char *
absolute_path(char *pathP, char *cdP)
{
	static char	full_path[2048];

	char	*P, *P1;
	int		slashes;

	P = full_path;
	if (*pathP == '/') {
		/* If the path is an absolute path we are done */
		strcpy(full_path, pathP);
	} else if (!cwdP || cwdP[0] != '/' || (cdP && *cdP == '/')) {
		/* Don't know what the cwd is  or cdP overrides it */
		if (cdP) {
			strcpy(full_path, cdP);
			P += strlen(cdP);
		}
		strcpy(P, pathP);
	} else {
		strcpy(P, cwdP);
		P += strlen(cwdP);
		*P++ = '/';
		if (cdP) {
			strcpy(P, cdP);
			P += strlen(cdP);
		}
		strcpy(P, pathP);
	}

//	fprintf(stderr, "pathP=%s cdP=%s -> %s\n", pathP, cdP, full_path);

	for (P = full_path; P1 = strchr(P, '.'); ) {
		switch (P1[1]) {
		case '/':
			if (P1 == full_path || P1[-1] == '/') {
				// Erase ./
				P = P1;
				strcpy(P, P+2);
				continue;
			}
			break;
		case '.':
			if (P1[2] == '/' && P1 > full_path+1 && P1[-1] == '/') {
				// Remove <parent>/../
				for (P = P1-2; P != pathP && P[-1] != '/'; --P);
				strcpy(P, P1+3);
				continue;
			}
			break;
		}
		P = P1 + 1;
	}
	return(full_path);
}

static int
call_cc(char *pathP, char *argv[], char * const env[])
{
	int		ret, pid;
	int		status;
	int		i;

	for (i = 0; argv[i]; ++i) {
		fprintf(stderr, "%s ", argv[i]);
	}
	fprintf(stderr,"\n");
	
	ret = fork();
	switch (ret) {
	case 0:
	{
		if (!use_stdin) {
   			char *P = strrchr(pathP,'/'); /* Switch to right directory        */
			if (P) {                      /* because gcc to dumb to do this   */
        		if (P == pathP) {         /* and want object in source dir    */
            		if (chdir("/")) {   
                		fprintf(stderr,"Can't change directory to /\n");
	                	error_exit(2);
					}
	        	}  else {    
            		*P = 0;            
            		if (chdir(pathP)) {     
                		fprintf(stderr,"Can't change directory to %s\n", pathP);
                		error_exit(2);
					}
					*P = '/';
       	}	} 	}
		execve(argv[0], argv, env);
		fprintf(stderr,"Return from execve %s\n", argv[0]);
		error_exit(1);
	}
	case -1:
		fprintf(stderr,"Fork to cc failed\n");
		error_exit(2);
	}
	pid = wait(&status);
	if (pid != ret) {
		fprintf(stderr,"Wait for cc failed\n");
		error_exit(3);
	}
	if (!WIFEXITED(status)) {
		fprintf(stderr,"cc did not terminate normally\n");
		return(-1);
	}
	return(WEXITSTATUS(status));
}

static void
dump_instances(void)
{
	Cdirectory	*directoryP;
	Carchive	*archiveP;
	Csource		*sourceP;
	Cclass		*classP;
	Cfunction	*functionP;
	Cvariable	*variableP;
	unsigned int	i, cnt;

	cnt = 0;
	for (directoryP = Cdirectory::g_headP; directoryP; directoryP = directoryP->m_nextP) {
		directoryP->emit_instance();
		++cnt;
	}
	if (cnt != Cdirectory::g_cnt) {
		fprintf(stderr, "!!Warning!! - Lost directories %d seen but %d created\n", cnt, Cdirectory::g_cnt);
	}

	cnt = 0;
	for (archiveP = Carchive::g_headP; archiveP; archiveP = archiveP->m_nextP) {
		archiveP->emit_instance();
		++cnt;
	}
	if (cnt != Carchive::g_cnt) {
		fprintf(stderr, "!!Warning!! - Lost archives %d seen but %d created\n", cnt, Carchive::g_cnt);
	}

	cnt = 0;
	for (sourceP = Csource::g_headP; sourceP; sourceP = sourceP->m_nextP) {
		sourceP->emit_instance();
		++cnt;
	}
	if (cnt != Csource::g_cnt) {
		fprintf(stderr, "!!Warning!! - Lost sources %d seen but %d created\n", cnt, Csource::g_cnt);
	}

	cnt = 0;
	if ((classP = Cclass::g_headP)) {
		fputs("$INSTANCE C C\n", stdout);
		for (; classP; classP = classP->m_nextP) {
			classP->emit_instance();
			++cnt;
	}	}
	if (cnt != Cclass::g_cnt) {
		fprintf(stderr, "!!Warning!! - Lost classes %d seen but %d created\n", cnt, Cclass::g_cnt);
	}

	cnt = 0;
	for (i = 0; i < HASHSIZE; ++i) {
		for (variableP = Cvariable::g_hash[i]; variableP; variableP = variableP->m_hashP) {
			if (!(variableP->m_flags & IgnoreX)) {
				variableP->emit_instance();
			}
			++cnt;
	}	}
	if (cnt != Cvariable::g_hash_cnt) {
		fprintf(stderr, "!!Warning!! - Lost variables %d seen but %d created\n", cnt, Cvariable::g_hash_cnt);
	}

	cnt = 0;
	for (i = 0; i < HASHSIZE; ++i) {
		for (functionP = Cfunction::g_hash[i]; functionP; functionP = functionP->m_hashP) {
			functionP->emit_instance();
			++cnt;
	}	}
	if (cnt != Cfunction::g_hash_cnt) {
		fprintf(stderr, "!!Warning!! - Lost functions %d seen but %d created\n", cnt, Cfunction::g_hash_cnt);
	}
}

static void
dump_edges(void)
{
	Cobject		*objectP;
	unsigned 	int cnt;

	cnt = 0;
	for (objectP = Cobject::g_created_headP; objectP; objectP = objectP->m_created_nextP) {
		objectP->emit_edges();
		++cnt;
	}
	if (cnt != Cobject::g_cnt) {
		fprintf(stderr, "!!Warning!! - Lost objects %d seen but %d created\n", cnt, Cobject::g_cnt);
	}

}

static void
dump_attributes(void)
{
	Cdirectory	*directoryP;
	Carchive	*archiveP;
	Csource		*sourceP;
	Cclass		*classP;
	Cvariable	*variableP;
	Cfunction	*functionP;
	Cedge		*edgeP;
	char		*demangleP;
	char		*nameP;
	unsigned int i;
	unsigned int cnt;

	fprintf(stdout, "\nFACT ATTRIBUTE :\n\n");

	cnt = 0;
	for (directoryP = Cdirectory::g_headP; directoryP; directoryP = directoryP->m_nextP) {
		directoryP->emit_attributes();
		++cnt;
	}
	if (cnt != Cdirectory::g_cnt) {
		fprintf(stderr, "!!Warning!! - Lost directories attributes %d seen but %d created\n", cnt, Cdirectory::g_cnt);
	}

	cnt = 0;
	for (archiveP = Carchive::g_headP; archiveP; archiveP = archiveP->m_nextP) {
		archiveP->emit_attributes();
		++cnt;
	}
	
	if (cnt != Carchive::g_cnt) {
		fprintf(stderr, "!!Warning!! - Lost archives attributes %d seen but %d created\n", cnt, Carchive::g_cnt);
	}

	cnt = 0;
	for (sourceP = Csource::g_headP; sourceP; sourceP = sourceP->m_nextP) {
		sourceP->emit_attributes();
		++cnt;
	}
	if (cnt != Csource::g_cnt) {
		fprintf(stderr, "!!Warning!! - Lost source attributes %d seen but %d created\n", cnt, Csource::g_cnt);
	}

	cnt = 0;
	if ((classP = Cclass::g_headP)) {
		fprintf(stdout, "C { label=\":classes\" }\n");
		for (; classP; classP = classP->m_nextP) {
			classP->emit_attributes();
			++cnt;
	}	}
	if (cnt != Cclass::g_cnt) {
		fprintf(stderr, "!!Warning!! - Lost class attributes %d seen but %d created\n", cnt, Cclass::g_cnt);
	}

	cnt = 0;
	for (i = 0; i < HASHSIZE; ++i) {
		for (variableP = Cvariable::g_hash[i]; variableP; variableP = variableP->m_hashP) {
			if (!(variableP->m_flags & IgnoreX)) {
				variableP->emit_attributes();
			}
			++cnt;
	}	}
	if (cnt != Cvariable::g_hash_cnt) {
		fprintf(stderr, "!!Warning!! - Lost variable attributes %d seen but %d created\n", cnt, Cvariable::g_hash_cnt);
	}

	cnt = 0;
	for (i = 0; i < HASHSIZE; ++i) {
		for (functionP = Cfunction::g_hash[i]; functionP; functionP = functionP->m_hashP) {
			functionP->emit_attributes();
			++cnt;
	}	}

	if (cnt != Cfunction::g_hash_cnt) {
		fprintf(stderr, "!!Warning!! - Lost function attributes %d seen but %d created\n", cnt, Cfunction::g_hash_cnt);
	}

	cnt = 0;
	for (i = 0; i < HASHSIZE; ++i) {
		for (edgeP = Cedge::g_hash[i]; edgeP; edgeP = edgeP->m_hashP) {
			edgeP->emit_attributes();
			++cnt;
	}	}
	if (cnt != Cedge::g_hash_cnt) {
		fprintf(stderr, "!!Warning!! - Lost edge attributes %d seen but %d created\n", cnt, Cedge::g_hash_cnt);
	}

}

static void
dump_schema(void)
{
	fputs(				"\nSCHEME TUPLE :\n\n"
		  				"$INHERIT D   $ENTITY\n"	// Directory
						"$INHERIT A   D\n"			// Archive
						"$INHERIT S   $ENTITY\n"	// Source
						"$INHERIT F   $ENTITY\n"	// Function
						"$INHERIT GF  F\n"			// GlobalFunction
						"$INHERIT EF  GF\n"			// ExternalFunction
						"$INHERIT LF  F\n", stdout);// LocalFunction
	if (show_variables != No_variables) {
		fputs(			"$INHERIT V   $ENTITY\n"	// Variable
						"$INHERIT GV  V\n"			// GlobalVariable
						"$INHERIT WV  GV\n"			// WeakGlobalVariable
						"$INHERIT EV  GV\n"			// ExternalVariable
						"$INHERIT SV  V\n"			// StaticVariable
						"$INHERIT FV  SV\n", stdout);// StaticFunctionVariable
	}	

	if (Cclass::g_headP) {
		fputs(			"$INHERIT C   $ENTITY\n", stdout);// Class
	}

	fputs(				"contain  D   D\n" 		// Directory contains Directory
		  				"contain  D   S\n"		// Directory contains Source
		  				"contain  S   F\n"		// Source    contains Functions
                        "contain  S   V\n", stdout);	// Source    contains Variables

	if (!no_function_calls) {
		fputs(			"CGF      F   GF\n"      // CallsGlobalFunction
						"CEF      F   EF\n"      // CallsExternalFunction
						"CLF      F   LF\n", stdout); // CallsLocalFunction
	}
	if (show_variables != No_variables) {
		fputs(			"UGV      F   GV\n"      // UsesGlobalVariable
						"UEV      F   EV\n"      // UsesExternalVariable
						"USV      F   SV\n"      // UsesStaticVariable
						"UFV      F   FV\n", stdout); // UsesStaticFunctionVariable
	}

	if (Cclass::g_headP) {
		fputs(			"M        C   F\n", stdout);	// Member of class is Function
		if (show_variables != No_variables) {
			fputs(		"M        C   V\n", stdout);	// Member of class is Variable
	}	}

	fputs(				"\nSCHEME ATTRIBUTE :\n\n"
		  				"$ENTITY {\n"
		  				"   color       = (255 255 255)\n"
          				"   labelcolor  = (  0   0   0)\n"
          				"   class_style = 1\n"
		  				"}\n"
						"D {\n"
                        "   class_label = \"Directory\"\n"
		  				"   color       = (255 255 153)\n"
          				"   labelcolor  = (  0   0   0)\n"
          				"   class_style = 4\n"
		  				"}\n"
						"A {\n"
						"   class_label = \"Library\"\n"
		  				"   color       = (153 255 102)\n"
          				"   labelcolor  = (  0   0   0)\n"
          				"   class_style = 4\n"
		  				"}\n"
						"S {\n"
						"   class_label = \"Source\"\n"
		  				"   color       = (  0 204   0)\n"
          				"   labelcolor  = (  0   0   0)\n"
          				"   class_style = 0\n"
		  				"}\n"
		  				"F {\n"
						"   class_label = \"Function\"\n"
          				"   class_style = 1\n"
		  				"}\n"
		  				"GF {\n"
						"   class_label = \"Global Function\"\n"
		  				"   color       = (255 102 102)\n"
          				"   labelcolor  = (  0   0   0)\n"
          				"   class_style = 1\n"
		  				"}\n"
						"EF {\n"
						"   class_label = \"External Function\"\n"
		  				"   color       = (255   0 204)\n"
          				"   labelcolor  = (  0   0   0)\n"
          				"   class_style = 1\n"
						"   class_description = \"Undeclared function\"\n"
		  				"}\n"
						"LF {\n"
						"   class_label = \"Local Function\"\n"
		  				"   color       = (255 153 153)\n"
          				"   labelcolor  = (  0   0   0)\n"
          				"   class_style = 1\n"
		  				"}\n", stdout);

	if (show_variables != No_variables) {
		fputs(			"V {\n"
						"   class_label = \"Variable\"\n"
                  		"   class_style = 6\n"
			      		"}\n"
						"GV {\n"
						"  class_label  = \"Global Variable\"\n"
                  		"  color       = (  0   0 255)\n"
                  		"  labelcolor  = (  0   0   0)\n"
                  		"  class_style = 6\n"
			      		"}\n"
						"WV {\n"
						"  class_label  = \"Weak Global Variable\"\n"
                  		"  color       = (204 204 255)\n"
                  		"  labelcolor  = (  0   0   0)\n"
                  		"  class_style = 6\n"
			      		"}\n"
						"EV {\n"
						"  class_label = \"External Variable\"\n"
                  		"  color       = (153  51 255)\n"
                  		"  labelcolor  = (  0   0   0)\n"
                  		"  class_style = 6\n"
						"  class_description = \"Undeclared variable\"\n"
			      		"}\n"
						"SV {\n"
						"  class_label = \"Static Variable\"\n"
                  		"  color       = (  0 204 204)\n"
                  		"  labelcolor  = (  0   0   0)\n"
                  		"  class_style = 6\n"
						"  class_description = \"File scope variable\"\n"
			      		"}\n"
						"FV {\n"
						"  class_label = \"Static Function Variable\"\n"
                      	"  color       = (102 255 255)\n"
                      	"  labelcolor  = (  0   0   0)\n"
                      	"  class_style = 6\n"
						"  class_description = \"Function scope variable\"\n"
			          	"}\n", stdout);
	}

	if (Cclass::g_headP) {
		fputs(			"C {\n"
						"   class_label = \"Class\"\n"
		      			"   color       = (255 204   0)\n"
              			"   labelcolor  = (  0   0   0)\n"
              			"   class_style = 1\n"
			          	"}\n", stdout);
	}

	if (!no_function_calls) {
		fputs(			"(CGF) {\n"
						"   class_label = \"Calls Global Function\"\n"
						"   color       = (153   0   0)\n"
						"   class_style = 0\n"
						"}\n"
						"(CEF) {\n"
						"   class_label = \"Calls External Function\"\n"
						"   color       = (153   0   0)\n"
						"   class_style = 0\n"
						"}\n"
						"(CLF)  {\n"
						"   class_label = \"Calls Local Function\"\n"
						"   color       = (153   0   0)\n"
						"   class_style = 0\n"
						"}\n", stdout);
	}
	if (show_variables != No_variables) {
		fputs(			"(UGV) {\n"
						"   class_label = \"Uses Global Variable\"\n"
						"   color       = (167 142  38)\n"
						"   class_style = 0\n"
						"}\n"
						"(UEV) {\n"
						"   class_label = \"Uses External Variable\"\n"
						"   color       = (167 142  38)\n"
						"   class_style = 0\n"
						"}\n"
						"(USV) {\n"
						"   class_label = \"Uses Static Variable\"\n"
						"   color       = (167 142 38)\n"
						"   class_style = 0\n"
						"}\n"
						"(UFV) {\n"
						"   class_label = \"Uses Static Function Variable\"\n"
						"   color       = (167 142  38)\n"
						"   class_style = 0\n"
						"}\n", stdout);
	}

	if (Cclass::g_headP) {
		fputs(			"(M) {\n"
						"   class_label = \"Member\"\n"
						"   color       = (255 255 255)\n"
	    				"   class_style = 0\n"
						"}\n", stdout);
	}

	fputs(				"\nFACT TUPLE :\n\n", stdout);
}

static void
collapse_directories(void)
{
	Cdirectory	*directoryP, *parentP;
	Cdirectory	*collapseP;

	for (directoryP = Cdirectory::g_headP; directoryP; directoryP = directoryP->m_nextP) {
		if ((parentP = (Cdirectory *) directoryP->m_parentP)) {
			parentP->m_collapse++;
	}	}
	
	collapseP = 0;
	for (directoryP = Cdirectory::g_headP; directoryP; directoryP = directoryP->m_nextP) {
		parentP = (Cdirectory *) directoryP->m_parentP;
		if (!parentP) {
			// No point collapsing a directory that is already the root
			continue;
		}
		for (;; parentP = (Cdirectory *) parentP->m_parentP) {
			if (!parentP) {
				if (!collapseP) {
					collapseP = directoryP;
//					fprintf(stderr, "Initial collapse %s\n", collapseP->m_nameP);
				}
				break;
			}
			if (parentP->m_collapse > 1) {
				break;
			}
			if (parentP == collapseP) {
				collapseP = directoryP;
//				fprintf(stderr, "New collapse %s\n", collapseP->m_nameP);
	}	}	}
		
	if (collapseP) {

		char		buffer[2048];
		int			lth;

		lth = collapseP->full_path(buffer);
		buffer[--lth] = 0;
		collapseP->name(buffer);

		for (parentP = collapseP; parentP = (Cdirectory *) parentP->m_parentP; ) {
			parentP->m_collapse = -2;
		}
		collapseP->m_parentP = 0;
}	}

static void
dump_ta(void)
{
#ifdef SANITY
	void sanity();

	fprintf(stderr, "Final sanity check\n");
	sanity();
#endif

	dump_schema();
	dump_instances();
	dump_edges();
	dump_attributes();
}

static int			read_object_suffix;
static int			read_object_seen_section;
static Csource		*read_object_sourceP;
static Csource		*read_object_executableP;
static Centity		*read_object_source_underP;
static Carchive		*read_object_archiveP;
static Cfunction	*read_object_functionP;
static Cfile		*read_object_fileP;
static int			read_object_lineno;
static int			read_object_seen_end;


static void
read_object_callback(int lth, char *inputP)
{
	Cfunction	*callingP;
	Cedge		*edgeP;
	char		*P, *P1, *P2, *absolute_pathP;
	int			c;

	if (!lth) {
		goto done;
	}
	switch (c = *inputP) {
	case 'A':
		if (!strncmp(inputP, "Archive ", 8)) {
			P              = inputP + 8;
			absolute_pathP = absolute_path(P, 0);
			read_object_archiveP = new Carchive(absolute_pathP);
			if (!read_object_archiveP) {
				outofmemory();
			}
			read_object_source_underP = read_object_archiveP;
		}
		goto done;
	case 'F':
		if (!strncmp(inputP, "File ", 5)) {
			P    = inputP + 5;
			lth -= 5;
			read_object_sourceP = new Csource(P, read_object_archiveP);
			if (!read_object_sourceP) {
				outofmemory();
			}
			goto done;
		}
		if (!strncmp(inputP, "Function ", 9)) {
			P = inputP + 9;
			goto is_function;
		}
		goto done;
	case 'C':
		if (!strncmp(inputP, "Calls ", 6)) {
			P = inputP + 6;
			goto is_call;
		}
		goto done;
	case 'D':
		if (!read_object_seen_section) {
			if (!strncmp(inputP, "Disassembly of section ", 23)) {
				if (!strcmp(inputP+23, ".init")) {
					/* This is an object or .so file */
					read_object_executableP = read_object_sourceP;
			}	}
			read_object_seen_section = 1;
		}
		goto done;
	case 'S':
		if (!strncmp(inputP, "Source ", 7)) {
			P              = inputP + 7;
			absolute_pathP = 0;
			if (read_object_executableP) {
				absolute_pathP = absolute_path(P, 0);
				read_object_sourceP = Csource::locate(absolute_pathP, read_object_executableP);
			}
			read_object_fileP = Cfile::locate(read_object_sourceP, 0);
			if (!read_object_fileP) {
				if (!absolute_pathP) {
					absolute_pathP = absolute_path(P, 0);
				}
				read_object_fileP = new Cfile(read_object_sourceP, 0, absolute_pathP);
				if (!read_object_fileP) {
					outofmemory();
			}	}
			read_object_lineno = 0;
		}
		goto done;
	case 'L':
		if (!strncmp(inputP, "Line ", 5)) {
			P = inputP + 5;
			read_object_lineno = strtol(P, &P1, 10);
		}
		goto done;
	case 'E':
		if (!strncmp(inputP, "End of report", 13)) {
			read_object_seen_end = 1;
		}
		goto done;
	}
	if (lth < 10) {
		goto done;
	}
	if (lth > 28 && inputP[8] == ':') {
		P = inputP + 32;
		if (!strncmp(P, "call ", 5)) {
			P = strchr(P, '<');
			if (!P) {
				goto done;
			}
			++P;
			P1 = strchr(P, '>');
			if (!P1) {
				goto done;
			}
			*P1 = 0;
			if (P1-P > 4) {
				P2 = P1 - 4;
				if (!strcmp(P2, "@PLT")) {
					*P2 = 0;
			}	}
is_call:
			callingP = Cfunction::locate(P, 0);
			if (!read_object_functionP) {
				/* Don't know how this could happen */
				goto done;
			}
			edgeP = Cedge::locate(read_object_functionP, callingP, read_object_lineno, read_object_fileP);
			edgeP->m_freq++;
		}
		goto done;
	}

	if ((c >= '0' && c <= '9') || (c >= 'a' && c <= 'f')) {
		Csource		*sourceP;
		Cfunction	*functionP;

		P = inputP + 9;
		if (*P++ != '<') {
			goto done;
		}
		P1 = strchr(P, '>');
		if (!P1) {
			goto done;
		}
		*P1 = 0;
is_function:
		read_object_functionP = Cfunction::locate(P, read_object_sourceP);
		if (!read_object_functionP->m_fileP) {
			read_object_functionP->m_fileP  = read_object_fileP;
			read_object_functionP->m_lineno = read_object_lineno;
		} 
		goto done;
	}
done:
	return;
}

static void
read_object_init(int suffix)
{
	read_object_suffix      = suffix;
	read_object_seen_section= 0;
	read_object_archiveP    = 0;
	read_object_executableP = 0;
	read_object_source_underP = 0;
	read_object_sourceP     = 0;
	read_object_fileP       = 0;
	read_object_lineno      = 0;
	read_object_functionP   = 0;
	read_object_seen_end    = 0;
}

static void
read_object_final(char *filenameP)
{
	if (!read_object_seen_end)  {
		fprintf(stderr, "Failed to fully process %s\n", filenameP);
		g_ret = 1;
}	}

static void
read_object(char *filenameP, int suffix)
{
	read_object_init(suffix);
	objdump_init(g_program_nameP);
	objdumpfile(filenameP, read_object_callback);
	objdump_finalize();
	read_object_final(filenameP);
}

static void
read_cx(char *filenameP, int suffix, char *argv[])
{
	read_object_init(suffix);
	cx(argv, read_object_callback);
	read_object_final(filenameP);
}

static	char	*inputP;
static  size_t	input_lth;

static char *
end_symbol_name(char *P)
{
	char	*P1;
	int		c;

	c = *P;
	if (c == '_' || isalpha(c)) {
		for (P1 = P; (c = *P1); ++P1) {
			if (!isalnum(c) && c != '_' && c != '$' && c != '.') {
				break;
		}	}
		return(P1);
	}
	return(0);
}

static void
read_asm(FILE *F, Csource *sourceP)
{
	Cfunction	*functionP;
	Cfile		*fileP;
	Cfunction	*callingP;
	Cvariable	*usesP;
	Cedge		*edgeP;
	int			lth;
	char		*P, *P1, *P2, *absolute_pathP;
	int			number;
	int			lineno;
	int			c, bracket, seen_global, seen_local, seen_weak, flags;

	fileP       = 0;
	lineno      = 0;
	functionP   = 0;
	seen_global = 0;
	seen_local  = 0;
	seen_weak   = 0;
	for (;;) {
		lth = getline(&inputP, &input_lth, F);
		if (lth < 0) {
			break;
		}
		switch (*inputP) {
		case '.':
			P = inputP + 1;
			if (!strncmp(P, "globl ", 6)) {		// .globl
				seen_global = 1;
			}
			continue;
		case '\t':			// Line starts with a tab
			P = inputP + 1;
			switch (*P) {
			case '.':
				++P;		// Line starts with '\t' '.'
				switch (*P) {
				case 'w':
					if (!strncmp(P, "weak\t", 5)) {	// \t.weak
						seen_weak = 1;
					}
					continue;
				case 'c':
					if (!strncmp(P, "comm\t", 5)) {	// \t.comm
						flags = 0;
						if (seen_local) {
							seen_local = 0;
							flags      = LocalX;
						}
						P += 5;
						P1 = end_symbol_name(P);
						if (P1) {
							P1[0] = 0;
							if (show_variables != No_variables) {
								usesP = Cvariable::locate(P, sourceP);
								usesP->m_flags |= flags;
					}	}	}
					continue;
				case 't':
					if (!strncmp(P, "type\t", 5)) {	
						flags = LocalX;
						if (seen_weak) {
							flags       = WeakX;
							seen_weak   = 0;
						}
						if (seen_global) {
							flags       = 0;
							seen_global = 0;
						}
						P += 5;
						P1 = strchr(P, ',');
						if (!P1) {
							continue;
						}
						if (!strncmp(P1, ", @object", 9)) {
							P1[0] = 0;
							if (show_variables != No_variables) {
								usesP = Cvariable::locate(P, sourceP);
								usesP->m_flags |= flags | InitX;
							}
							continue;
						}
						if (!strncmp(P1, ", @function", 11)) {
							P1[0] = 0;
							functionP = Cfunction::locate(P, sourceP);
							functionP->m_flags |= flags;
						}
					}
					continue;
				case 'f':
					if (!strncmp(P, "file ", 5)) {
						P += 5;
						number = strtol(P, &P1, 10);
						if (P == P1) {
							continue;
						}
						if (*P1 != ' ') {
							continue;
						}
						++P1;
						if (*P1 != '"') {
							continue;
						}
						P = P1 + 1;
						P1 = strchr(P, '"');
						if (!P1) {
							continue;
						}
						*P1 = 0;
	
						absolute_pathP = absolute_path(P, cwd1P);
						fileP = new Cfile(sourceP, number, absolute_pathP);
						if (!fileP) {
							outofmemory();
						}
					}
					continue;
				case 'l':
					if (!strncmp(P, "loc", 3)) {
						P += 3;
						if (*P == ' ') {	// .loc
							number = strtol(P, &P1, 10);
							if (P == P1) {
								continue;
							}
							fileP = Cfile::locate(sourceP, number);
							if (!fileP) {
								// This should never happen
								continue;
							}
							P = P1;
							lineno = strtol(P, &P1, 10);
							if (P == P1) {
								continue;
							}
							if (functionP && !functionP->m_fileP) {
								functionP->m_fileP = fileP;
								functionP->m_lineno = lineno;
							}
							continue;
						}
						if (!strncmp(P, "al\t", 3)) {	// .local
							seen_local = 1;
					}	}
				}
				continue;
			case 'c':
				if (no_function_calls) {
					continue;
				}
				if (!strncmp(P, "call\t", 5)) {
					P += 5;
					P1 = strchr(P, '\n');
					if (!P1) {
						continue;
					}
					*P1 = 0;
					if (P1-P > 4) {
						P2 = P1-4;
						if (!strcmp(P2, "@PLT")) {
							*P2 = 0;
					}	}
					callingP = Cfunction::locate(P, 0);
					if (functionP) {
						edgeP = Cedge::locate(functionP, callingP, lineno, fileP);
						edgeP->m_freq++;
					}
					continue;
				}
			default:
				if (show_variables == No_variables) {
					continue;
				}
				P = strchr(P, '\t');
				if (!P) {
					continue;
				}
/*
 http://www.kernel-traffic.org/wine/wn20000228_32.html

 The linker/loader maintain a global offset table, which consists of an array
 of pointers to exported variables. All access to those variables is supposed
 to go through these pointers, if the variable could reside within another
 dynamic object. 

 If 'variable' is an exported symbol, then 'variable@GOT' is replaced by the
 linker with a constant: the offset of the GOT slot containing the pointer to
 variable. So, if the %ebx register contains the start address of the GOT, you
 can retrieve this pointer via 

 movl variable@GOT(%ebx), %eax 

 If you want the value of the variable, you need another indirection: 

 movl (%eax), %eax 

 If you know, however, that the variable in question must reside in the same
 dynamic object as the user, then you can avoid that double indirection, by
 using the GOTOFF facility. 'variable@GOTOFF' is replaced by the linker with
 another constant: the difference between the absolute address of the variable
 and the address of the global offset table (of the current dynamic object).
 Thus, you can retrieve the value of the variable in one step using: 

 movl variable@GOTOFF(%ebx), %eax 

*/
				++P;
				if (*P == '$') {	// Address of
					++P;
				}
				P1 = end_symbol_name(P);

				if (P1) {
					*P1 = 0;
					usesP = Cvariable::locate(P, 0);
					if (functionP) {
						edgeP = Cedge::locate(functionP, usesP, lineno, fileP);
						edgeP->m_freq++;
					}
					*P1 = c;
					P   = P1;
				}
				// Search for ','
				bracket = 0;
				for (;; ++P) {
					switch (*P) {
					case ',':
						if (bracket) {
							continue;
						}
					case 0:
						break;
					case '(':
						++bracket;
						continue;
					case ')':
						--bracket;
					default:
						continue;
					}
					break;
				}

				if (*P != ',') {
					continue;
				}
				for (; (c = *++P) && isspace(c); );
				if (c == '$') {
					c = *++P;
				}
				P1 = end_symbol_name(P);
				if (P1) {
					*P1 = 0;
					usesP = Cvariable::locate(P, 0);
					if (functionP) {
						edgeP = Cedge::locate(functionP, usesP, lineno, fileP);
						edgeP->m_freq++;
				}	}
}	}	}	}	

int
main(int argc, char *argv[], char * const env[])
{
	static char	*argv1[512] =
	{ 
		0,
		"-S",
		"-g",
		0,
		0
	};

	char		*asx_compileP, *asx_unlinkP, *asx_ignoreP, *asx_forceP;
	char		*P, *P1, *P2, *P3, *nextP, *suffixP, *absolute_pathP;
	int 		c, i, j, k, lth, arg, arg1, seen_file, unlink_me, flags;
	FILE		*F;
	Cdirectory	*directoryP;
	Csource		*sourceP;
	Cfunction	*functionP;
	Cvariable	*variableP;
	int			ret, unlink_flag, compile_flag, special_arg;
	char		*stdin_bufferP;
	size_t		stdin_buffer_lth;
	size_t		stdin_lth;
	char		filename1[1024];
	char		*assemblerP;
	int			lineno;
	clock_t		start_clock, end_clock;
	
	start_clock = time(0);

	compile_flag = 1;
	unlink_flag  = 1;

	printf("// ASX Version 1.0.7 Compiled " __DATE__ ":"  __TIME__ " Start: %s\n", ctime(&start_clock));

	asx_compileP = 0;
	asx_unlinkP  = 0;
	asx_ignoreP  = 0;
	asx_forceP   = 0;

	for (i = j = 1; P = argv[i]; ++i) {
		if (!strncmp(P, "--ASX_COMPILE=", 14)) {
			asx_compileP = P + 14;
			continue;
		}
		if (!strncmp(P, "--ASX_UNLINK=", 13)) {
			asx_unlinkP = P + 13;
			continue;
		}
		if (!strncmp(P, "--ASX_IGNORE=", 13)) {
			asx_ignoreP = P + 13;
			continue;
		}
		if (!strncmp(P, "--ASX_FORCE=", 12)) {
			asx_forceP = P + 12;
			continue;
		}
		argv[j++] = P;
	}
	argv[j] = 0;
	argc    = j;

	if (!asx_compileP) {
		asx_compileP = getenv("ASX_COMPILE");
	}
	if (asx_compileP) {
		printf("// ASX_COMPILE=%s// Will not compile source code", asx_compileP);
		if (!strcasecmp(asx_compileP, "no")) {
			printf(" (Extract by directly examining source code)\n");
			compile_flag = 0;
			unlink_flag  = 0;
		} else {
			printf(" (Extract by compiling source code to assembler)\n");
	}	}
		
	if (compile_flag) {
		if (!asx_unlinkP) {
			asx_unlinkP = getenv("ASX_UNLINK");
		}
		if (asx_unlinkP) {
			printf("// ASX_UNLINK=%s", asx_unlinkP);
			if (!strcasecmp(asx_unlinkP, "no")) {
				unlink_flag = 0;
				printf(" (Generated *.s files preserved)\n");
			} else {
				printf(" (Generated *.s files removed)\n");
	}	}	}

	if (!asx_ignoreP) {
		asx_ignoreP = getenv("ASX_IGNORE");
	}
	if (asx_ignoreP) {
		printf("// ASX_IGNORE=%s\n", asx_ignoreP);

		for (P = asx_ignoreP; P; P = (c ? P1+1 : 0)) {
			P1 = P;
			c  = *P;
			// Remove leading space
			if (isspace(c)) {
				continue;
			}
			for (; (c = *P1) && c != ';' && c != ':' && c != ','; ++P1);
			*P1 = 0;
			// Remove trailing space
			for (P2 = P1; --P2 >= P; ) {
				if (!isspace(*P2)) {
					break;
				}
				*P2 = 0;
			}
			switch (*P) {
			case 'f':
			case 'F':
				no_function_calls = 1;
				continue;
			case 'v':
			case 'V':
				switch (P[1]) {
				case 's':
				case 'S':
					show_variables = Global_variables;
					continue;
				case 'f':
				case 'F':
					show_variables = Static_variables;
					continue;
				}
				show_variables = No_variables;
				continue;
			case 0:
				// Ignore things that only contained whitespace
				continue;
			}

			for (k = sizeof(suffixes)/sizeof(suffixes[0]); ; ) {
				if (--k < 0) {
					fprintf(stderr, "Can't ignore suffix \"%s\" - unknown\n", P);
					break;
				}
				suffixP = suffixes[k];
				if (!suffixP || strcmp(P, suffixP)) {
					continue;
				}
				suffixes[k] = 0;
				break;
			}
	}	}

	if (!asx_forceP) {
		asx_forceP = getenv("ASX_FORCE");
	}
	if (asx_forceP) {
		printf("// ASX_FORCE=%s\n", asx_forceP);
		for (P = asx_forceP; P; P = (c ? P1+1 : 0)) {
			P1 = P;
			c  = *P;
			// Remove leading space
			if (isspace(c)) {
				continue;
			}
			for (; (c = *P1) && c != ';' && c != ':' && c != ','; ++P1);
			*P1 = 0;
			// Remove trailing space
			for (P2 = P1; --P2 >= P; ) {
				if (!isspace(*P2)) {
					break;
				}
				*P2 = 0;
			}
			switch (*P) {
			case 'v':
			case 'V':
				switch(P[1]) {
				case 'g':
				case 'G':
					show_variables = Global_variables;
					continue;
				case 'S':
					show_variables = Static_variables;
					continue;
				}
				show_variables = Static_function_variables;
				continue;
			case 0:
				// Ignore things that only contained whitespace
				continue;
			}

			for (k = sizeof(suffixes)/sizeof(suffixes[0]); ; ) {
				if (--k < 0) {
					fprintf(stderr, "Can't force suffix \"%s\" - unknown\n", P);
					break;
				}
				suffixP = suffixes[k];
				if (!suffixP || strcmp(P, suffixP)) {
					continue;
				}
				force[k] = 1;
				break;
	}	}	}

	printf("// Arguments to %s:\n//", argv[0]);

	for (i = 1; argv[i]; ++i) {
		fputc(' ', stdout);
		fputs(argv[i], stdout);
	}
	fputc('\n', stdout);

	input_lth  = 120;
	inputP     = (char *) malloc(input_lth);
	if (!inputP) {
		fprintf(stderr, "Can't malloc input buffer\n");
		exit(1);
	}

	g_program_nameP = argv[0];
	g_ret           = 0;

	arg         = 3;
	seen_file   = 0;
	special_arg = 0;
	if (argc < 2 || !strcmp(argv[1], "-")) {
    	use_stdin = 1;
		stdin_buffer_lth = 120;
		stdin_bufferP    = (char *) malloc(stdin_buffer_lth);
		if (!stdin_bufferP) {
			fprintf(stderr, "Can't malloc stdin buffer\n");
			exit(1);
		}
	} else {
		cwdP = getcwd(0, 0);
	}

	lineno = 0;

	for (i = j = 1; ; ) {
		if (use_stdin) {
			if (use_stdin == 1) {
next_stdin:
				arg = 3;
		        if (getline(&stdin_bufferP, &stdin_buffer_lth, stdin) < 0) {
            		break;
        		}
				++lineno;
				P = stdin_bufferP;
				if (*P != '"') {
					fprintf(stderr, "**Line %d: Missing cwd start '\"' >%s\n", lineno, P);
					goto next_stdin;
				}
				cwdP  = P+1;
				nextP = strchr(cwdP, '"');
				if (!nextP) {
					fprintf(stderr, "**Line %d: Missing cwd end '\" >%s'\n", lineno, cwdP);
					goto next_stdin;
				}
				*nextP++ = 0;
				if (chdir(cwdP)) {
					fprintf(stderr,"Can't change directory to %s\n", cwdP);
					goto next_stdin;
				}
				P    = strchr(nextP, '"');
				if (!P) {
					fprintf(stderr, "**Line %d: Missing program start '\"' >%s\n", lineno, nextP);
					goto next_stdin;
				}
				++P;
				nextP = strchr(P, '"');
				if (!nextP) {
					fprintf(stderr, "**Line %d: Missing program end '\"' >%s\n", lineno, P);
					goto next_stdin;
				}
				*nextP++ = 0;
				argv1[0] = P;
				use_stdin = 2;
			}
			P = strchr(nextP, '"');
			if (!P) {
				/* No more arguments */
				goto next_stdin;
			}
			++P;
			for (nextP = P1 = P; ; ++nextP) {
				switch (c = *nextP) {
				case 0:
					fprintf(stderr, "**Line %d: Missing argument end '\"' >%s\n", lineno, P);
					goto next_stdin;
				case '"':
					goto have_token;
				case '\\':
					c = *(++nextP);
				}
				*P1++ = c;
			}
have_token:
			*P1 = 0;
			++nextP;
		} else {
			if (i >= argc) {
				break;
			}
			P = argv[i++];			/* so can write out source.s		*/
		}
		switch (special_arg) {
		case  1:
			argv1[arg++] = P;
		case -1:
			special_arg = 0;
			continue;
		}

		if (P[0] == '-') {
			if (seen_file) {
				arg       = 3;
				seen_file = 0;
			}
			switch (P[1]) {
			case 'a':
				if (!strcmp(P, "-aux-info")) {
					special_arg = 1;
				}
				break;
			case 'i':
				special_arg = 1;
				break;
			case 'm':
			case 'D':
			case 'U':
			case 'I':
				if (!P[2]) {
					special_arg = 1;
				}
				break;
			case 'o':
				if (!P[2]) {
					special_arg = -1;
				}
			case 'g':
			case 'c':
			case 'S':
			case 'O':
				continue;
			case 'E':
			case 'M':
				if (P[2] == 'T' || P[2] == 'F') {
					special_arg = -1;
					continue;
				} 
				if (P[2] == 'D' || P[2] == 'P') {
					continue;
				}
				if (use_stdin) {
					/* If the compiler was not compiling we shouldn't */
					fprintf(stderr, "**Line %d: Unknown compiler option '%s'\n", lineno, P);
					goto next_stdin;
				}
				continue;
			}
			argv1[arg++] = P;
			continue;
		}

		if (!strcmp(P, "/dev/null")) {
			goto skip;
		}

		P1 = strrchr(P, '.');
		if (!P1) {
			k = SUFFIX_none;
		} else {
			++P1;
			for (k = sizeof(suffixes)/sizeof(suffixes[0]); --k >= 0; ) {
				suffixP = suffixes[k];
				if (suffixP && !strcmp(P1, suffixP)) {
					break;
		}	}	}
		assemblerP = P;		/* This is the default when suffix is *.s */
		unlink_me  = 0;
		switch (k) {
		case SUFFIX_so:
		case SUFFIX_o:
			if (use_stdin && !force[k]) {
				int		s;
				char	*suffixP;
			
				lth = P1 - P;
				strncpy(filename1, P, lth);

				if (k == SUFFIX_so) {
					s = SUFFIX_o;
				} else {
					s = SUFFIX_S;
				}
				for (; s < SUFFIX_none; ++s) {
					suffixP = suffixes[s];
					if (!suffixP) {
						continue;
					}
					strcpy(filename1+lth, suffixP);
					absolute_pathP = absolute_path(filename1, 0);
					if (Csource::lookup(absolute_pathP)) {
						fprintf(stderr, "Assuming %s generated from %s\n", P, absolute_pathP);
						goto skip;
			}	}	}
		case SUFFIX_none:
		case SUFFIX_a:
			seen_file  = 1;
			read_object(P, k);
			break;
		case SUFFIX_c:
		case SUFFIX_C:
		case SUFFIX_cc:
			if (!use_stdin) {
				argv1[0] = "/usr/bin/gcc";
			}
			goto generate_asm;
		case SUFFIX_cxx:
		case SUFFIX_cpp:
		case SUFFIX_cpp1:
			if (!use_stdin) {
				argv1[0] = "/usr/bin/g++";
			}
generate_asm:
			seen_file = 1;
			if (use_stdin) {
				P2 = P;
			} else {
				P2 = strrchr(P, '/');
				if (!P2) {
					P2 = P;
				} else {
					++P2;
			}	}	
			lth = P1-P;
			memcpy(filename1, P, lth);
			strcpy(filename1+lth, "s");
			assemblerP = filename1;
			arg1 = arg;
			if (use_stdin) {
				argv1[arg1++] = "-o";
				argv1[arg1++] = assemblerP;
			}
			argv1[arg1++] = P2;
			argv1[arg1] = 0;
	
			if (compile_flag) {
				if (unlink_flag && access(assemblerP, F_OK)) {
					unlink_me = 1;
				}

				ret = call_cc(P, argv1, env);
				if (!ret) {
					if (!use_stdin) {
						if (P2 != P) {
							strcpy(cwd1, P);
							cwd1[P2-P] = 0;
							cwd1P = cwd1;
					}	}
				} else {
					printf("// Failed to compile %s\n", P);
				}
			} else {
				ret = -1;
			}
			if (ret) {
				argv1[arg] = absolute_path(P, 0);
				read_cx(P, k, argv1);
				unlink_me = 0;
				continue;
			}
		case SUFFIX_S:
			seen_file      = 1;
			absolute_pathP = absolute_path(P, 0);
			sourceP        = new Csource(absolute_pathP, 0);
			if (!sourceP) {
				outofmemory();
			}
	
			F = fopen(assemblerP, "r");
			if (!F) {
				fprintf(stderr, "Can't open %s\n", assemblerP);
				continue;
			}
			read_asm(F, sourceP);
			fclose(F);
			cwd1P = 0;

			if (!unlink_me) {
				break;
			}

			unlink_me = 0;

			/* Sanity checks follow */
			if (assemblerP != filename1) {
				fprintf(stderr, "May not unlink file %s\n", assemblerP);
				exit(5);
				break;
			} 
			lth = strlen(assemblerP);
			if (lth<3 || assemblerP[lth-2] != '.' || assemblerP[lth-1] != 's') {
				fprintf(stderr, "May not delete %s\n", assemblerP);
				exit(5);
				break;
			}

			/* Discard this assembly file when done */

			if (unlink(assemblerP)) {
				fprintf(stderr, "Can't delete %s\n", assemblerP);
				break;
			}
/*
			fprintf(stderr, "Deleted %s\n", assemblerP);
 */
			break;
		}
skip:
		continue;
	}

	sourceP = 0;
	for (i = 0; i < HASHSIZE; ++i) {
		for (variableP = Cvariable::g_hash[i]; variableP; variableP = variableP->m_hashP) {
			P     = variableP->m_nameP;
			flags = variableP->m_flags;
			if (flags & LocalX) {
				P1 = strchr(P, '.');
				if (P1) {
					/* The assembler labels local static variables declared
					 * within functions with a numeric suffix to
					 * avoid collisions with global variables and the
					 * same named local variables in other functions.
					 */
					flags              |= NestedX;
					variableP->m_flags  = flags;
					*P1 = 0;
			}	}
			switch (*P) {
			case '_':
				switch (P[1]) {
				case 'Z':
					if (P[2] == 'Z' && (flags & LocalX)) {
						flags              |= NestedX;
						variableP->m_flags  = NestedX;
					}
					break;
				case '_':
					if (!strcmp(P,"__PRETTY_FUNCTION__")) {
						variableP->m_flags |= IgnoreX;
						continue;
					}
					break;
			}	}

			switch (show_variables) {
			case Global_variables:
				if (flags & LocalX) {
					variableP->m_flags |= IgnoreX;
					continue;
				}
				break;
			case Static_variables:
				if (flags & NestedX) {
					variableP->m_flags |= IgnoreX;
					continue;
			}	}
			if (!variableP->m_parentP) {
				if (!sourceP) {
					sourceP    = new Csource(":globals:", 0);
					if (!sourceP) {
						outofmemory();
					}
				}
				variableP->m_parentP = sourceP;
				variableP->m_flags  |= ExternalX;
				variableP->add_created();
			}
	}	}

	sourceP = 0;
	for (i = 0; i < HASHSIZE; ++i) {
		for (functionP = Cfunction::g_hash[i]; functionP; functionP = functionP->m_hashP) {
			if (!functionP->m_parentP) {
				if (!sourceP) {
					sourceP    = new Csource(":library:", 0);
					if (!sourceP) {
						outofmemory();
					}
				}
				functionP->m_parentP = sourceP;
				functionP->m_flags  |= ExternalX;
				functionP->add_created();
			}
	}	}
	collapse_directories();
	dump_ta();
	end_clock = time(0);
	printf("\n// End: %s \n", ctime(&end_clock));
	return(g_ret);
}

