#include <malloc.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <stdlib.h>

#include "util.h"
#include "object.h"
#include "classmember.h"
#include "class.h"
#include "directory.h"
#include "source.h"
#include "file.h"
#include "function.h"

#include "demangle.h"

Cfunction	*Cfunction::g_hash[HASHSIZE] = {0};
int			Cfunction::g_hash_cnt        = 0;

/*
static void trap(void)
{
}
*/

Cfunction::Cfunction(const char *nameP) : Cclassmember(nameP)
{
	m_hashP       = 0;

/*
	{
		int lth = strlen(nameP);
		if (lth > 4 && !strcmp(nameP+lth-4, "@PLT")) {
			trap();
		}
	}
*/
}

Cfunction *
Cfunction::locate(const char *nameP, Centity *parentP)
{
	unsigned int	index;
	Cfunction		**functionPP, *functionP, *sameP;
	Centity			*parent1P;

	sameP         = 0;
	index         = hash(nameP);
	for (functionPP = g_hash + index; functionP = *functionPP; functionPP = &(functionP->m_hashP)) {
		if (!strcmp(functionP->m_nameP, nameP)) {
			if (parentP) {
				// This function must occur in this source
				// Otherwise it is a distinct declaration
				// Perhaps all the declarations are static ones
				parent1P = functionP->m_parentP;
				if (!parent1P) {
					/* function seen but location not yet known
                     * make this the parent of this function
                     */
					functionP->m_parentP = parentP;
					functionP->add_created();
				} else if (parent1P != parentP) {
					sameP = functionP;
					continue;
			}	}
			return(functionP);
	}	}
	functionP = new Cfunction(nameP);
	if (!functionP) {
		outofmemory();
	}
	/* Add to end so always return first function declaration seen if
	 * searching for location of the function declaration:
	 *
     * eg. if we are looking for foo() and foo() was in foo.c later
     *     also seen in foo.o return reference to foo.c:foo() in
	 *     preferance to foo.o:foo()
     */
	if (parentP) {
		functionP->m_parentP = parentP;
		functionP->add_created();
	}
	if (sameP) {
		functionP->m_sameP = sameP->m_sameP;
		sameP->m_sameP     = functionP;
	}
	*functionPP = functionP;
	++g_hash_cnt;
	return(functionP);
}

/* virtual */ objectE
Cfunction::type(void) const
{
	return(functionE);
}

/* virtual */ void
Cfunction::emit_id(void) const
{
	fprintf(stdout, "F%p", this);
}

const char *
Cfunction::get_type(void) const
{
	int			flags;
	const char 	*typeP;

	flags = m_flags;
	if (flags & LocalX) {
		typeP = "LF";
	} else if (flags & ExternalX) {
		typeP = "EF";
	} else {
		typeP = "GF";
	}
	return(typeP);
}

/* virtual */ void
Cfunction::emit_edges(void)
{
	Cclassmember::emit_edges();
	if (Cclass::g_headP) {
		emit_class();
	}
}
		
/* virtual */ void
Cfunction::emit_instance(void) const
{
	fprintf(stdout, "$INSTANCE F%p %s\n", this, get_type());
}

void
Cfunction::emit_attributes(void) const
{
	const char 	*nameP;
	char		*demangleP;

	emit_id();
	fputs(" {", stdout);

	nameP     = m_nameP;
	demangleP = 0;
	if (nameP[0] == '_') {
		demangleP = cplus_demangle(nameP, DMGL_ANSI|DMGL_PARAMS|DMGL_GNU_V3);
		if (demangleP) {
			fprintf(stdout,"mangle=\"%s\" ", nameP);
			nameP = demangleP;
	}	}
	fprintf(stdout, "label=\"%s\"", nameP);
	if (demangleP) {
		free(demangleP);
	}
	if (m_fileP) {
		char	*filenameP = m_fileP->m_nameP;
		char 	path_to_me[2048];

		path_to_me[0] = 0;
		full_path(path_to_me);

		// TODO: Improve this logic using ../ etc.
		if (!strcmp(filenameP, path_to_me)) {
			filenameP = ".";
		}
		fprintf(stdout, " file=\"%s\" lineno=%d", filenameP, m_lineno);
	}
	fputs("}\n", stdout);
}

/* virtual */ void
Cfunction::print(void) const
{
	fputs(m_nameP, stdout);
}
