#include <malloc.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <stdlib.h>

#include "util.h"
#include "object.h"
#include "classmember.h"
#include "class.h"
#include "directory.h"
#include "source.h"
#include "variable.h"

#include "demangle.h"

Cvariable	*Cvariable::g_hash[HASHSIZE] = {0};
int			Cvariable::g_hash_cnt        = 0;

Cvariable::Cvariable(const char *nameP) : Cclassmember(nameP)
{
	m_hashP     = 0;
}

Cvariable *
Cvariable::locate(const char *nameP, Centity *parentP)
{
	unsigned int	index;
	Cvariable		**variablePP, *variableP, *sameP;
	Centity			*parent1P;

	sameP         = 0;
	index         = hash(nameP);
	for (variablePP = g_hash + index; variableP = *variablePP; variablePP = &(variableP->m_hashP)) {
		if (!strcmp(variableP->m_nameP, nameP)) {
			if (parentP) {
				// This variable must occur in this source
				// Otherwise it is a distinct declaration
				// Perhaps all the declarations are static ones
				parent1P = variableP->m_parentP;
				if (!parent1P) {
					/* variable seen but location not yet known
                     * make this the parent of this variable
                     */
					variableP->m_parentP = parentP;
					variableP->add_created();
				} else if (parent1P != parentP) {
					sameP = variableP;
					continue;
			}	}
			return(variableP);
	}	}
	variableP = new Cvariable(nameP);
	if (!variableP) {
		outofmemory();
	}
	/* Add to end so always return first function declaration seen if
	 * searching for location of the function declaration:
	 *
     * eg. if we are looking for foo and foo was in foo.c later
     *     also seen in foo.o return reference to foo.c:foo in
	 *     preferance to foo.o:foo
     */
	if (parentP) {
		variableP->m_parentP = parentP;
		variableP->add_created();
	}
	if (sameP) {
		variableP->m_sameP = sameP->m_sameP;
		sameP->m_sameP     = variableP;
	}
	*variablePP = variableP;
	++Cvariable::g_hash_cnt;
	return(variableP);
}

/* virtual */ objectE
Cvariable::type(void) const
{
	return(variableE);
}

void
trap_variable(void)
{
	fprintf(stderr, "Seen\n");
}

/* virtual */ void
Cvariable::emit_id(void) const
{
	if (((void *) this) == ((void *) 0xa189200)) {
		trap_variable();
	}
	fprintf(stdout, "V%p", this);
}

const char *
Cvariable::get_type(void) const
{
	int			flags;
	const char 	*typeP;

	flags = m_flags;
	if (flags & NestedX) {
		typeP = "FV";
	} else if (flags & LocalX) {
		typeP = "SV";
	} else if (flags & WeakX) {
		typeP = "WV";
	} else if (flags & ExternalX) {
		typeP = "EV";
	} else {
		typeP = "GV";
	}
	return(typeP);
}

/* virtual */ void
Cvariable::emit_edges(void)
{
	if (!(m_flags & IgnoreX)) {
		Cclassmember::emit_edges();
		if (Cclass::g_headP) {
			emit_class();
	}	}
}
		
/* virtual */ void
Cvariable::emit_instance(void) const
{
	fprintf(stdout, "$INSTANCE V%p %s\n", this, get_type());
}

void
Cvariable::emit_attributes(void) const
{
	const char	*nameP;
	char		*demangleP;

	emit_id();
	fputs(" {", stdout);

	nameP     = m_nameP;
	demangleP = 0;
	if (nameP[0] == '_') {
		demangleP = cplus_demangle(nameP, DMGL_ANSI|DMGL_PARAMS|DMGL_GNU_V3);
		if (demangleP) {
			fprintf(stdout, "mangle=\"%s\" ", nameP);
			nameP = demangleP;
	}	}
	fprintf(stdout, "label=\"%s\" file=\".\"", nameP);
	if (demangleP) {
		free(demangleP);
	}
	fputs("}\n", stdout);
}

/* virtual */ void
Cvariable::print(void) const
{
	fputs(m_nameP, stdout);
}
